"""Endpoints for the GTI"""

ENDPOINT_INIT = "gti/public/init"
ENDPOINT_CHECK_NAME = "gti/public/checkName"
ENDPOINT_GET_ROUTE = "gti/public/getRoute"
ENDPOINT_DEPARTURE_LIST = "gti/public/departureList"
ENDPOINT_GET_TARIFF = "gti/public/getTariff"
ENDPOINT_DEPARTURE_COURSE = "gti/public/departureCourse"
ENDPOINT_LIST_STATIONS = "gti/public/listStations"
ENDPOINT_LIST_LINES = "gti/public/listLines"
ENDPOINT_GET_ANNOUNCEMENTS = "gti/public/getAnnouncements"
ENDPOINT_GET_INDIVIDUAL_ROUTE = "gti/public/getIndividualRoute"
ENDPOINT_GET_VEHICLE_MAP = "gti/public/getVehicleMap"
ENDPOINT_GET_TRACK_COORDINATES = "gti/public/getTrackCoordinates"
ENDPOINT_CHECK_POSTAL_CODE = "gti/public/checkPostalCode"
ENDPOINT_GET_STATION_INFORMATION = "gti/public/getStationInformation"
ENDPOINT_TARIFF_ZONE_NEIGHBOURS = "gti/public/tariffZoneNeighbours"
ENDPOINT_TICKET_LIST = "gti/public/ticketList"



import asyncio
import os
from datetime import datetime, timedelta

import aiohttp
from pygti.auth import Auth
from pygti.gti import GTI

try:
    from dotenv import load_dotenv

    load_dotenv()

    GTI_USER = os.getenv("GTI_USER")
    GTI_PASS = os.getenv("GTI_PASS")
except ImportError:
    pass

if not (GTI_USER and GTI_PASS):
    print("To run the examples, enter your credentials for the GTI API.")
    GTI_USER = input("GTI Username: ")
    GTI_PASS = input("GTI Password: ")


async def main():
    async with aiohttp.ClientSession() as session:

        auth = Auth(session, GTI_USER, GTI_PASS)
        gti = GTI(auth)

        print("Example 1: init()")
        ir = await gti.init()
        print(ir)

        print()
        print("Example 2: checkName()")
        cn = await gti.checkName({"theName": {"name": "Wartenau"}})
        print(cn)

        print()
        print("Example 3: getRoute()")
        payload = {
            "language": "de",
            "start": {
                "name": "Ritterstraße",
                "city": "Hamburg",
                "combinedName": "Ritterstraße",
                "id": "Master:60904",
                "type": "STATION",
                "coordinate": {"x": 10.046196, "y": 53.567617},
            },
            "dest": {
                "name": "Wartenau",
                "city": "Hamburg",
                "combinedName": "Wartenau",
                "id": "Master:10901",
                "type": "STATION",
                "coordinate": {"x": 10.035515, "y": 53.56478},
            },
            "time": {"date": "heute", "time": "jetzt"},
            "timeIsDeparture": True,
            "realtime": "REALTIME",
        }
        gr = await gti.getRoute(payload)
        print(gr)

        print()
        print("Example 4.1: departureList()")
        dl = await gti.departureList(
            {
                "station": {
                    "name": "Wartenau",
                    "id": "Master:10901",
                    "type": "STATION",
                },
                "time": {"date": "heute", "time": "jetzt"},
                "maxList": 5,
                "maxTimeOffset": 200,
                "useRealtime": True,
            }
        )

        print(dl)

        print()
        print("Example 4.2: departureList(), return filters")
        dl = await gti.departureList(
            {
                "station": {
                    "name": "Wartenau",
                    "id": "Master:10901",
                    "type": "STATION",
                },
                "time": {"date": "heute", "time": "jetzt"},
                "maxList": 5,
                "maxTimeOffset": 200,
                "useRealtime": True,
                "returnFilters": True,
            }
        )
        print(dl)

        print()
        print("Example 5: getTariff()")
        gT = await gti.getTariff(
            {
                "scheduleElements": [
                    {
                        "departureStationId": "Master:10950",
                        "arrivalStationId": "Master:37979",
                        "lineId": "DB-EFZ:RE8_DB-EFZ_Z",
                    }
                ],
                "departure": {"date": "16.02.2020", "time": "8:04"},
                "arrival": {"date": "16.02.2020", "time": "8:29"},
            }
        )
        print(gT)

        print("Example 7: listStations()")
        # used a older dataReleaseID to show changes since then in the response
        ls = await gti.listStations({"dataReleaseID": "32.17.02"})
        print(ls)

        print()
        print("Example 8: listLines()")
        ll = await gti.listLines({"dataReleaseID": "32.17.02"})
        print(ll)

        print("Example 9: AnnouncementRequest")
        ar = await gti.getAnnouncements(
            {
                "names": ["S3"],
                "timeRange": {
                    "begin": datetime.now() - timedelta(days=2),
                    "end": datetime.now() + timedelta(days=10),
                },
            }
        )

        print(ar)

        print()
        print("Example 10: getIndividualRoute()")
        payload = {
            "starts": [
                {"type": "ADDRESS", "coordinate": {"x": 9.92496, "y": 53.563494}}
            ],
            "dests": [
                {"type": "ADDRESS", "coordinate": {"x": 9.924269, "y": 53.562925}}
            ],
            "maxLength": 10000,
            "serviceType": "BICYCLE",
            "profile": "BICYCLE_NORMAL",
            "speed": "NORMAL",
        }
        indRoute = await gti.getIndividualRoute(payload)
        print(indRoute)

        print()
        print("Example 11: getVehicleMap()")
        payload = {
            "boundingBox": {
                "lowerLeft": {"x": 9.985707, "y": 53.573138, "type": "EPSG_4326"},
                "upperRight": {"x": 9.992702, "y": 53.576916, "type": "EPSG_4326"},
            },
            "periodBegin": int(datetime.timestamp(datetime.now())),
            "periodEnd": int(datetime.timestamp(datetime.now() + timedelta(minutes=1))),
            "withoutCoords": True,
            "coordinateType": "EPSG_31467",
            "vehicleTypes": ["U_BAHN"],
            "realtime": False,
        }
        vm = await gti.getVehicleMap(payload)
        print(vm)

        print()
        print("Example 12: getTrackCoordinates()")
        payload = {
            "coordinateType": "EPSG_4326",
            "stopPointKeys": [
                "ZVU-DB:8004248:2",
                "ZVU-DB:8004247:2",
                "ZVU-DB:809100:1",
                "ZVU-DB:119106:1",
            ],
        }
        tc = await gti.getTrackCoordinates(payload)
        print(tc)

        print()
        print("Example 13: checkPostalCode()")
        pc = await gti.checkPostalCode({"postalCode": 20355})
        print(pc)

        print()
        print("Example 14: stationInformation()")
        si = await gti.stationInformation(
            {"station": {"name": "Wartenau", "id": "Master:10901", "type": "STATION"}}
        )
        print(si)

        print()
        print("Example 15: tariffZoneNeighbours()")
        tzn = await gti.tariffZoneNeighbours({})
        print(tzn)

        print()
        print("Example 16: ticketList()")
        tl = await gti.ticketList({"stationKey": "Master:92903"})
        print(tl)


asyncio.run(main())



class GTIError(Exception):
    """An exception occured while using the GTI API."""

    def __init__(self, return_code, error_text, error_dev_info):
        self.return_code = return_code
        self.error_text = error_text
        self.error_dev_info = error_dev_info


class CannotConnect(Exception):
    """Exception raised if connection could not be established to host."""

    pass


class InvalidAuth(GTIError):
    """Exception raised if credentials are incorrent."""

    pass


class CheckNameTooMany(GTIError):
    """Check Name returned too many hits.

    Precise search term."""

    pass


class CommunicationError(GTIError):
    """Error while connecting to backend."""

    pass


class RouteError(GTIError):
    """Route could not be calculated."""

    pass



from .auth import Auth
from .const import *
from .schemas import (
    AnnouncementRequest,
    CNRequest,
    DLRequest,
    GRRequest,
    IndividualRouteRequest,
    LLRequest,
    LSRequest,
    PostalCodeRequest,
    SIRequest,
    TariffRequest,
    TariffZoneNeighboursRequest,
    TLRequest,
    TrackCoordinatesRequest,
    VehicleMapRequest,
)


class GTI:
    def __init__(self, auth):
        self.auth = auth

    async def init(self):
        response = await self.auth.request("post", ENDPOINT_INIT, {})
        return await response.json()

    async def checkName(self, payload):
        request = CNRequest(payload)
        response = await self.auth.request("post", ENDPOINT_CHECK_NAME, request)
        return await response.json()

    async def getRoute(self, payload):
        request = GRRequest(payload)
        response = await self.auth.request("post", ENDPOINT_GET_ROUTE, request)
        return await response.json()

    async def departureList(self, payload):
        request = DLRequest(payload)
        response = await self.auth.request("post", ENDPOINT_DEPARTURE_LIST, request)
        return await response.json()

    async def getIndividualRoute(self, payload):
        request = IndividualRouteRequest(payload)
        response = await self.auth.request(
            "post", ENDPOINT_GET_INDIVIDUAL_ROUTE, request
        )
        return await response.json()

    async def listLines(self, payload):
        request = LLRequest(payload)
        response = await self.auth.request("post", ENDPOINT_LIST_LINES, request)
        return await response.json()

    async def getTariff(self, payload):
        request = TariffRequest(payload)
        response = await self.auth.request("post", ENDPOINT_GET_TARIFF, request)
        return await response.json()

    async def listStations(self, payload):
        request = LSRequest(payload)
        response = await self.auth.request("post", ENDPOINT_LIST_STATIONS, request)
        return await response.json()

    async def getVehicleMap(self, payload):
        request = VehicleMapRequest(payload)
        response = await self.auth.request("post", ENDPOINT_GET_VEHICLE_MAP, request)
        return await response.json()

    async def getTrackCoordinates(self, payload):
        request = TrackCoordinatesRequest(payload)
        response = await self.auth.request(
            "post", ENDPOINT_GET_TRACK_COORDINATES, request
        )
        return await response.json()

    async def stationInformation(self, payload):
        request = SIRequest(payload)
        response = await self.auth.request(
            "post", ENDPOINT_GET_STATION_INFORMATION, request
        )
        return await response.json()

    async def getAnnouncements(self, payload):
        request = AnnouncementRequest(payload)
        response = await self.auth.request("post", ENDPOINT_GET_ANNOUNCEMENTS, request)
        return await response.json()

    async def checkPostalCode(self, payload):
        request = PostalCodeRequest(payload)
        response = await self.auth.request("post", ENDPOINT_CHECK_POSTAL_CODE, request)
        return await response.json()

    async def tariffZoneNeighbours(self, payload):
        request = TariffZoneNeighboursRequest(payload)
        response = await self.auth.request(
            "post", ENDPOINT_TARIFF_ZONE_NEIGHBOURS, request
        )
        return await response.json()

    async def ticketList(self, payload):
        request = TLRequest(payload)
        response = await self.auth.request("post", ENDPOINT_TICKET_LIST, request)
        return await response.json()



from datetime import datetime

import pytz
from voluptuous import All, In, Length, LengthInvalid, Range, Required, Schema

CoordinateType = In(["EPSG_4326", "EPSG_31467"])

SDType = In(["STATION", "COORDINATE", "ADDRESS", "POI", "UNKNOWN"])

Language = In(["de", "en"])

FilterType = In(["HVV_Listed", "NO_FILTER"])


class ModLength(Length):
    def __init__(self, min=None, max=None, msg=None, mod=None):
        super(self.__class__, self).__init__(min=min, max=max, msg=msg)
        self.mod = mod

    def __call__(self, v):
        super(self.__class__, self).__call__(v)
        if self.mod is not None and len(v) % self.mod != 0:
            raise LengthInvalid(
                self.msg or "length of value must be divisible by %s" % self.mod
            )
        return v


def DateTime(dt):
    dt = pytz.utc.localize(dt)
    return f"{dt.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]}{dt.strftime('%z')}"


# See 1.4 in gti docs
BaseRequestType = Schema({"language": Language, "filterType": FilterType})

FilterServiceType = In(
    [
        "ZUG",
        "UBAHN",
        "SBAHN",
        "AKN",
        "RBAHN",
        "FERNBAHN",
        "BUS",
        "STADTBUS",
        "METROBUS",
        "SCHNELLBUS",
        "NACHTBUS",
        "EILBUS",
        "AST",
        "FAEHRE",
    ]
)

AnnouncementFilterPlannedType = In(["ONLY_PLANNED", "ONLY_UNPLANNED", "NO_FILTER"])

TariffDetails = Schema(
    {
        "innerCity": bool,
        "cityTraffic": bool,
        "gratis": bool,
        "greaterArea": bool,
        "tariffZones": [int],
        "counties": [str],
        "rings": [str],
        "fareStage": bool,
        "fareStageNumber": int,
        "tariffNames": [str],
    }
)

Coordinate = Schema({"x": float, "y": float, "type": CoordinateType})

SDName = Schema(
    {
        "name": str,
        "city": str,
        "combinedName": str,
        "id": str,
        "type": SDType,
        "coordinate": Coordinate,
        "tariffDetails": TariffDetails,
        "serviceTypes": [str],
        "hasStationInformation": bool,
    }
)

CNRequest = Schema.extend(
    BaseRequestType,
    {
        "theName": SDName,
        "maxList": int,
        "maxDistance": int,
        "coordinateType": CoordinateType,
        "tariffDetails": bool,
        "allowTypeSwitch": bool,
    },
)

GTITime = Schema({"date": str, "time": str})

FilterEntry = Schema(
    {Required("serviceID"): str, "stationIDs": [str], "serviceName": str, "label": str}
)


DLRequest = Schema.extend(
    BaseRequestType,
    {
        "station": SDName,
        "stations": [SDName],
        "time": GTITime,
        "maxList": int,
        "maxTimeOffset": int,
        "allStationsInChangingNode": bool,
        "returnFilters": bool,
        "filter": [FilterEntry],
        "serviceTypes": [FilterServiceType],
        "useRealtime": bool,
    },
)

SIRequest = Schema.extend(BaseRequestType, {"station": SDName})

ContSearchByServiceId = Schema(
    {
        "serviceId": int,
        "lineKey": str,
        "plannedDepArrTime": GTITime,
        "additionalOffset": int,
    }
)

# kinds: GTI Docs say int but traffic on the website uses strings
# Same thing with the value of the penalty. All are strings
TariffInfoSelector = Schema({"tariff": str, "tariffRegions": bool, "kinds": [str]})
PenaltyName = In(
    [
        "ChangeEvent",
        "ExtraFare",
        "Walker",
        "AnyHandicap",
        "ToStartStationBy",
        "TimeRange",
        "ForVisitors",
        "desiredType",
        "DesiredLine",
    ]
)

Penalty = Schema({"name": PenaltyName, "value": str})

RealtimeType = In(["REALTIME", "PLANDATA", "AUTO"])

SimpleServiceType = In(
    [
        "BUS",
        "TRAIN",
        "SHIP",
        "FOOTPATH",
        "BICYCLE",
        "AIRPLANE",
        "CHANGE",
        "CHANGE_SAME_PLATFORM",
    ]
)

ServiceType = Schema(
    {"simpleType": SimpleServiceType, "shortInfo": str, "longInfo": str, "model": str}
)
GRRequest = Schema.extend(
    BaseRequestType,
    {
        Required("start"): SDName,
        Required("dest"): SDName,
        "via": SDName,
        "time": GTITime,
        "timeIsDeparture": bool,
        "numberOfSchedules": int,
        "tariffDetails": bool,
        "continousSearch": bool,
        "contSearchByServiceId": ContSearchByServiceId,
        "coordinateType": CoordinateType,
        "schedulesBefore": int,
        "schedulesAfter": int,
        "returnReduced": bool,
        "tariffInfoSelector": [TariffInfoSelector],
        "penalties": [Penalty],
        "returnPartialTickets": bool,
        "realtime": RealtimeType,
        "intermediateStops": bool,
        "useStationPosition": bool,
        "forcedStart": SDName,
        "forcedDest": SDName,
        "toStartBy": SimpleServiceType,
        "toDestBy": SimpleServiceType,
        "returnContSearchData": bool,
    },
)

LineModificationType = In(["MAIN", "SEQUENCE"])

LLRequest = Schema.extend(
    BaseRequestType,
    {
        "dataReleaseID": str,
        "modificationTypes": [LineModificationType],
        "withSublines": bool,
    },
)

ScheduleElementLight = Schema(
    {"departureStationId": str, "arrivalStationId": str, "lineId": str}
)

TariffRequest = Schema.extend(
    BaseRequestType,
    {
        Required("scheduleElements"): [ScheduleElementLight],
        Required("departure"): GTITime,
        Required("arrival"): GTITime,
        "returnReduced": bool,
        "returnPartialTickets": bool,
    },
)

IndividualProfileType = In(
    [
        "BICYCLE_NORMAL",
        "BICYCLE_RACING",
        "BICYCLE_QUIET_ROADS",
        "BICYCLE_MAIN_ROADS",
        "BICYCLE_BAD_WEATHER",
        "FOOT_NORMAL",
    ]
)

IndividualRouteRequest = Schema.extend(
    BaseRequestType,
    {
        "starts": [SDName],
        "dests": [SDName],
        "maxLength": int,
        "maxResults": int,
        "type": CoordinateType,
        "serviceType": SimpleServiceType,
        "profile": IndividualProfileType,
        "speed": str,
    },
)

ModificationType = In(["MAIN", "POSITION"])

LSRequest = Schema.extend(
    BaseRequestType,
    {
        "dataReleaseID": str,
        "modificationTypes": [ModificationType],
        "coordinateType": CoordinateType,
        "filterEquivalent": bool,
    },
)

BoundingBox = Schema({"lowerLeft": Coordinate, "upperRight": Coordinate})

VehicleType = In(
    [
        "REGIONALBUS",
        "METROBUS",
        "NACHTBUS",
        "SCHNELLBUS",
        "EILBUS",
        "AST",
        "SCHIFF",
        "U_BAHN",
        "S_BAHN",
        "R_BAHN",
        "A_BAHN",
        "F_BAHN",
    ]
)

VehicleMapRequest = Schema.extend(
    BaseRequestType,
    {
        "boundingBox": BoundingBox,
        "periodBegin": int,
        "periodEnd": int,
        "withoutCoords": bool,
        "coordinateType": CoordinateType,
        "vehicleTypes": [VehicleType],
        "realtime": bool,
    },
)

SIRequest = Schema({"station": SDName})

TimeRange = Schema({Required("begin"): DateTime, Required("end"): DateTime})

AnnouncementRequest = Schema.extend(
    BaseRequestType,
    {
        "names": [str],
        "timeRange": TimeRange,
        "full": bool,
        "filterPlanned": AnnouncementFilterPlannedType,
    },
)

PostalCodeRequest = Schema.extend(
    BaseRequestType,
    {
        "postalCode": Range(
            min=1000,
            max=99999,
            msg="value must be a valid german postal code (5 digits)",
        )
    },
)

TrackCoordinatesRequest = Schema.extend(
    BaseRequestType,
    {
        "coordinateType": CoordinateType,
        Required("stopPointKeys"): All([str], ModLength(min=2, mod=2)),
    },
)

TariffZoneNeighboursRequest = BaseRequestType

TLRequest = Schema.extend(BaseRequestType, {"stationKey": str})



import base64
import hashlib
import hmac
import json

from aiohttp import ClientConnectorError, ClientResponse, ClientSession

from .exceptions import *

GTI_DEFAULT_HOST = "api-prod.geofox.de"


class Auth:
    """Class to make authenticated requests to the geofox api."""

    def __init__(
        self,
        websession: ClientSession,
        username: str,
        password: str,
        host: str = GTI_DEFAULT_HOST,
    ):
        """Initialize the auth."""
        self.websession = websession
        self.username = username
        self.password = password
        self.host = host

    async def request(
        self, method: str, path: str, payload=None, **kwargs
    ) -> ClientResponse:
        """Make a request."""
        headers = kwargs.get("headers")

        if headers is None:
            headers = {}
        else:
            headers = dict(headers)

        payload.update({"version": 38})

        data = json.dumps(payload).encode("UTF-8")

        signature = base64.b64encode(
            hmac.new(self.password.encode("UTF-8"), data, hashlib.sha1).digest()
        ).decode("UTF-8")

        headers["geofox-auth-signature"] = f"{signature}"
        headers["geofox-auth-user"] = self.username

        try:
            response = await self.websession.request(
                method,
                f"https://{self.host}/{path}",
                json=payload,
                **kwargs,
                headers=headers,
            )

            data = await response.json()

            return_code = data.get("returnCode")
            error_text = data.get("errorText")
            error_dev_info = data.get("errorDevInfo")

            if return_code == "ERROR_CN_TOO_MANY":
                raise CheckNameTooMany(return_code, error_text, error_dev_info)
            elif return_code == "ERROR_COMM":
                raise CommunicationError(return_code, error_text, error_dev_info)
            elif return_code == "ERROR_ROUTE":
                raise RouteError(return_code, error_text, error_dev_info)
            elif return_code == "ERROR_TEXT":
                if error_dev_info == "Authentication failed!":
                    raise InvalidAuth(return_code, error_text, error_dev_info)
                raise GTIError(return_code, error_text, error_dev_info)

            return response
        except ClientConnectorError as error:
            raise CannotConnect(error)
